import math
#Version: Got water drawage to work properly. But: Electrical energy is provided to all nodes.
class ewhModel:
    
    #Physical constants
    rho = 1000    #Denisity of water
    c = 4184      #Specific heat capacity of water [joule/(kg*Kelin)]

    def __init__(self, thermal_resistance, tank_volume):
        
        self.R = thermal_resistance
        #if(tank_volume == 200):
            #self.R = thermal_resistance * 0.6666
        self.TANK_LENGTH = 1
        #if(tank_volume == 200):self.TANK_LENGTH = 1.333
        self.TANK_VOLUME = tank_volume
        self.TANK_RADIUS = math.sqrt((self.TANK_VOLUME/1000)/(math.pi*self.TANK_LENGTH))
        self.TANK_AREA = 2*math.pi*self.TANK_RADIUS*self.TANK_LENGTH + 2*math.pi*self.TANK_RADIUS*self.TANK_RADIUS
        
        
        #print('len %d, vol %d, area %.3f, radius %.3f, R %.2f'%(self.TANK_LENGTH, 
        #                                                        self.TANK_VOLUME, 
        #                                                        self.TANK_AREA, 
        #                                                        self.TANK_RADIUS, 
        #                                                        self.R))
        
    
    # ---------------------------- Physical equations --------------------------------------
    def __waterEnthalpy__(self, water_temperature, ref_temperature, liters_water):
        #print("Enthalpy")
        return self.c * self.rho *(liters_water * 0.001) * (water_temperature - ref_temperature)
        
        
    def __deltaTemperature__(self, energy, liters_water):
        #print("Delta_temp")
        return energy/(self.c * self.rho * (liters_water * 0.001))
        
        
        
    def __thermalDecay__(self, time, t_inital, t_ambient, volume, thermal_resistance):
        #print("Decay")
        if(self.TANK_VOLUME == 200):
            return t_ambient + (t_inital - t_ambient)*math.exp((-1.0 * time)/(self.c*self.rho*(0.001*volume*0.75)*thermal_resistance))
        else:
            return t_ambient + (t_inital - t_ambient)*math.exp((-1.0 * time)/(self.c*self.rho*(0.001*volume)*thermal_resistance))
    
    # -------------------------------------------------------------------------------------------
    


class ewhModel_one(ewhModel):
    
    
    
    def __init__(self, thermal_resistance, tank_volume, t_initial):
        ewhModel.__init__(self, thermal_resistance, tank_volume) 
        self.t_inside_rst = t_initial
        self.t_inside = t_initial
        self.model_type = 'one'

    def reset(self):
        self.t_inside = self.t_inside_rst
        
    # ------------------------ Sim methods ---------------------------------------------
    def stepTime(self, time_sec, added_power_kw):
        #Increase due to power added
        element_energy_added = (added_power_kw*1000)*time_sec  #Convert to Watt and integrate
        self.t_inside += self.__deltaTemperature__(element_energy_added, self.TANK_VOLUME)
        
        #Decrease due to thermal losses
        self.t_inside = self.__thermalDecay__(time_sec, self.t_inside, self.t_amb, self.TANK_VOLUME, self.R);
        
    def stepVolume(self, volume_litres):
        #print("Stepping %.2f litres"%volume_litres)
        self.t_inside = ((self.TANK_VOLUME - volume_litres)/self.TANK_VOLUME) * (self.t_inside - self.t_inlet) + self.t_inlet
    # ------------------------------------------------------------------------------------
    
    
    # ------------------------ State getters and setters --------------------------------
    def getOutletTemp(self):
        return self.t_inside    #The one-node model assumes uniform temperature
        
    def setInletTemp(self, temp_degC):
        self.t_inlet = temp_degC
        
    def setAmbTemp(self, temp_degC):
        self.t_amb = temp_degC
    #------------------------------------------------------------------------------------
   



class ewhModel_two(ewhModel):
    
    node_state = False
    V_THRESHOLD = 2
    d_thermocline = 0.01;
    THERMOCLINE_MAGIC = 0.4
    r_upper = float()
    r_lower = float()
    v_upper = float()
    v_lower = float()
    r_lower = float()
    r_upper = float()
    r_thermocline = float()
    
    
    def __init__(self, thermal_resistance, tank_volume, t_initial, orient):
        ewhModel.__init__(self, thermal_resistance, tank_volume) 
        self.t_inside_rst = t_initial
        self.t_upper = self.t_lower = t_initial
        self.v_threshold_detect = 0
        self.model_type = 'two'
        self.ORIENT = orient
        self.v_lower = self.TANK_VOLUME
    
    def reset(self):
        self.t_upper = self.t_lower = self.t_inside_rst
        self.v_threshold_detect = 0
        self.node_state = False
    
    def stepTime(self, time_sec, added_power_kw):
        element_energy_added = (added_power_kw*1000)*time_sec  #Convert to Watt and integrate
        
        if(not self.node_state):
            #print('Time: 1')
            self.t_lower = self.__thermalDecay__(time_sec, self.t_lower, self.t_amb, self.TANK_VOLUME, self.R)
            self.t_lower += self.__deltaTemperature__(element_energy_added, self.TANK_VOLUME)
            self.t_upper = self.t_lower
        elif(self.node_state):
            #print('Time: 2')
            ## Outward radiation
            self.t_lower = self.__thermalDecay__(time_sec, self.t_lower, self.t_amb, self.v_lower, self.r_lower)
            self.t_upper = self.__thermalDecay__(time_sec, self.t_upper, self.t_amb, self.v_upper, self.r_upper)
            
            ## Cross-thermocline radiation
            t_lower_init = self.t_lower
            t_upper_init = self.t_upper
            self.t_lower = self.__thermalDecay__(time_sec, t_lower_init, t_upper_init, self.v_lower, self.r_thermocline)
            self.t_upper = self.__thermalDecay__(time_sec, t_upper_init, t_lower_init, self.v_upper, self.r_thermocline)
            
            ## Element temperature added
            self.t_lower += self.__deltaTemperature__(element_energy_added, self.v_lower)
            
            # Check for 1 node transition
            if(self.t_lower >= self.t_upper):
                #print('Time: Going 1 node')
                #self.t_upper += self.__deltaTemperature__(element_energy_added, self.TANK_VOLUME)
                self.t_lower = self.t_upper
                self.node_state = False
                #print('To one-node1(',added_power_kw, ')') 
                self.v_threshold_detect = 0
                
                self.v_lower = self.TANK_VOLUME
                self.v_upper = 0
            
    def stepVolume(self, volume_litres):
        
        # Check for 2 node transition
        init_temp = self.t_lower
        if(not self.node_state):
            if(volume_litres > 0):
                #print('Vol: Going 2 node')
                self.v_lower = 0
                self.t_lower = self.t_inlet
                self.v_upper = self.TANK_VOLUME
                self.v_threshold_detect = 0
                self.node_state = True
                #print('To two-node')
        #if(volume_litres > 0):
        #    tank_before = self.c*self.v_upper*0.001*self.t_upper + self.c*self.v_lower*0.001*self.t_lower
                
        if(not self.node_state):
            #print('Vol: 1')
            self.t_lower = ((self.TANK_VOLUME - volume_litres)/self.TANK_VOLUME) * (self.t_lower - self.t_inlet) + self.t_inlet
            self.t_upper = self.t_lower
        elif(self.node_state):
            #print('Vol: 2')
            ## Check for 1 node transition
            if((self.v_upper - volume_litres) <= 0):
                #print('Vol: Going 1 node')
                self.t_upper = self.t_lower
                self.node_state = False
                #print('To one-node2')
                self.v_threshold_detect = 0
                
                self.v_lower = self.TANK_VOLUME
                self.v_upper = 0
                
            else:
                self.v_lower += volume_litres
                self.v_upper -= volume_litres

                #Update lower temperature
                self.t_lower = ((self.v_lower - volume_litres)/self.v_lower) * (self.t_lower - self.t_inlet) + self.t_inlet

                #Update thermal condutances (node angle changed due to volume)
                self.updateConductances()
        #if(volume_litres > 0):
        #    print('tank used:', self.c*volume_litres*0.001*(self.t_upper-20))
        #    tank_after = self.c*self.v_upper*0.001*self.t_upper + self.c*self.v_lower*0.001*self.t_lower 
        #    print('tank diff:', tank_before - tank_after)
        
    # --------------------- Conductance calculations --------------------------
    
    def updateConductances(self):
        
        if (self.ORIENT is 'h'):
            #Caculate node angle using Newton-raphson
            node_angle = self.getNodeAngle();

            volume_m3 = self.v_lower * 0.001;
            
            # Calculate areas for Horizontal using node angle
            a_lower = 2*(volume_m3/self.TANK_LENGTH) + self.TANK_RADIUS*node_angle*self.TANK_LENGTH;
            a_upper = self.TANK_AREA - a_lower
            a_thermocline = 2*self.TANK_RADIUS*math.sin(node_angle/2)*self.TANK_LENGTH
            
        else:
            #Calculate areas for Vertical (much easier than horizontal)
            
            # area_lower = (1 side-area) + ((total manifold-area) * (lower_vol/total_volume))
            a_lower = math.pi*self.TANK_RADIUS*self.TANK_RADIUS + (2*math.pi*self.TANK_RADIUS*self.TANK_LENGTH)*(self.v_lower/self.TANK_VOLUME)
            a_upper = self.TANK_AREA - a_lower
            a_thermocline = math.pi*self.TANK_RADIUS*self.TANK_RADIUS
            
            
        # Calculate thermal resistances from area fractions
        self.r_lower = self.R*(self.TANK_AREA/a_lower)
        self.r_upper = self.R*(self.TANK_AREA/a_upper)
        self.r_thermocline = self.THERMOCLINE_MAGIC*a_thermocline
            
        
    # -------------------- Newton Raphson calculations for Horz ------------------------
    # See Philip's thesis "Rethinking Electric water Heaters"
    
    def getNodeAngle(self):
        stop_threshold = math.pi/100000;
        
        # -------- Raphson -----------------
        theta = math.pi #Initial guess
        while(abs(self.nrf(theta)) > stop_threshold):
            theta = theta - (self.nrf(theta)/self.nrf_prime(theta))
        # ----------------------------------
        return theta
    
    
    def nrf(self, theta):
        volume_m3 = self.v_lower*0.001
        return ((2*volume_m3)/(math.pow(self.TANK_RADIUS, 2)*self.TANK_LENGTH)) - theta + math.sin(theta)
    
    def nrf_prime(self, theta):
        return -1 + math.cos(theta)
    # --------------------------------------------------------------------------
    

class ewhModel_three(ewhModel):
    
    d_thermocline = 0.01;
    THERMOCLINE_MAGIC = 0.4
    node_state = False
    
    
    def __init__(self, thermal_resistance, tank_volume, t_initial, orient, nodes):
        ewhModel.__init__(self, thermal_resistance, tank_volume) 
        self.t_inside_rst = t_initial
        self.temperatures = [t_initial]*nodes
        self.v_threshold_detect = 0
        self.model_type = 'multi'
        self.ORIENT = orient
        self.volume = self.TANK_VOLUME/nodes
        self.nodes = nodes
        self.r_tank, self.r_thermocline = self.setupConductances()
        #print('r_tank:', self.r_tank, ' r_thermo:', self.r_thermocline)
    
    def reset(self):
        self.temperatures = [self.t_inside_rst]*nodes 
    
    def stepTime(self, time_sec, added_power_kw):
        
        ## Element temperature added
        element_energy_added = (added_power_kw*1000)*time_sec  #Convert to Watt and integrate
        self.temperatures[0] += self.__deltaTemperature__(element_energy_added, self.volume)
        for k in range(self.nodes-1):
            if(self.temperatures[k] > self.temperatures[k+1]):
                average_temperature = (self.temperatures[k]*(k+1) + self.temperatures[k+1])/(2+k)
                for i in range(k+2):
                    self.temperatures[i] = average_temperature

        ## Outward radiation
        init_temperatures = []
        for k in range(self.nodes):
            if(self.node_state == True):
                self.temperatures[k] = self.__thermalDecay__(time_sec, self.temperatures[k], self.t_amb, self.volume, self.r_tank[k])
            else:
                self.temperatures[k] = self.__thermalDecay__(time_sec, self.temperatures[k], self.t_amb, self.volume, self.R*self.nodes)
                   
            init_temperatures.append(self.temperatures[k])
            
        ## Cross-thermocline radiation
        for k in range(self.nodes):
            if(k == 0):
                self.temperatures[k] = self.__thermalDecay__(time_sec, self.temperatures[k], init_temperatures[k+1], self.volume, self.r_thermocline[k])
            elif(k == self.nodes-1):
                self.temperatures[k] = self.__thermalDecay__(time_sec, self.temperatures[k], init_temperatures[k-1], self.volume, self.r_thermocline[k])
            else:
                temp_diff = self.__thermalDecay__(time_sec, init_temperatures[k], init_temperatures[k-1], self.volume, self.r_thermocline[k]) - init_temperatures[k]
                self.temperatures[k] = self.__thermalDecay__(time_sec, init_temperatures[k], init_temperatures[k+1], self.volume, self.r_thermocline[k]) + temp_diff

        for k in range(self.nodes-1):
            if(self.temperatures[k] >= self.temperatures[k+1]):
                self.temperatures[k+1] = self.temperatures[k]

        ##Check transition to single-node
        if(all(round(x,1) == round(self.temperatures[0],1) for x in self.temperatures) and self.node_state == True):
            self.node_state = False
            
    def stepVolume(self, volume_litres):
        if(volume_litres > 0 and self.node_state == False):
            self.node_state = True
        
        #tank_before = 0
        #for k in range(self.nodes):
        #    tank_before += self.c*self.volume*0.001*self.temperatures[k]
        
        if(volume_litres > 0):
            cold_nodes = int(volume_litres/self.volume)
            remaining_cold = (volume_litres%self.volume)
            
            init_temperatures = []
            for k in range(self.nodes):
                init_temperatures.append(self.temperatures[k])


            for k in range(self.nodes):
                if(k == 0):
                    self.temperatures[k] = ((self.volume - remaining_cold)/self.volume) * (self.temperatures[k] - self.t_inlet) + self.t_inlet
                else:
                    self.temperatures[k] = ((self.volume - remaining_cold)/self.volume) * (self.temperatures[k] - init_temperatures[k-1]) + init_temperatures[k-1]
                        
            init_temperatures = []
            for k in range(self.nodes):
                init_temperatures.append(self.temperatures[k])
            for k in range(self.nodes):
                if(k < cold_nodes):
                    self.temperatures[k] = self.t_inlet
                else:
                    self.temperatures[k] = init_temperatures[k-cold_nodes]
         
            #tank_after = 0
            #for k in range(self.nodes):
            #    tank_after += self.c*self.volume*0.001*self.temperatures[k]     
            #print('tank diff:', tank_before - tank_after)

    # --------------------- Conductance calculations --------------------------

    def setupConductances(self):
        
        if (self.ORIENT is 'h'):
            
            a_previous = 0
            r_tank = []
            r_thermocline = []
            
            for i in range(int(self.nodes / 2)):
                iter_volume = self.volume * (i+1)
                node_angle = self.getNodeAngle(iter_volume)

                volume_m3 = iter_volume * 0.001;

                a_tank = 2*(volume_m3/self.TANK_LENGTH) + self.TANK_RADIUS*node_angle*self.TANK_LENGTH
                
                a_thermocline = 2*self.TANK_RADIUS*math.sin(node_angle/2)*self.TANK_LENGTH
                r_tank.append(self.R*(self.TANK_AREA/(a_tank - a_previous)))
                r_thermocline.append(self.R*(self.TANK_AREA/a_thermocline))
                a_previous = a_tank
                
            for j in range(int(self.nodes / 2)):
                r_tank.append(r_tank[-1 - j*2])
                r_thermocline.append(r_thermocline[-1 - j*2])
                
            
        else:
            
            a_bound = math.pi*self.TANK_RADIUS*self.TANK_RADIUS + (2*math.pi*self.TANK_RADIUS*self.TANK_LENGTH)*(self.volume/self.TANK_VOLUME)
            a_middle = (2*math.pi*self.TANK_RADIUS*self.TANK_LENGTH)*(self.volume/self.TANK_VOLUME)
            a_thermocline = math.pi*self.TANK_RADIUS*self.TANK_RADIUS

            #r_bound = self.R*(self.TANK_AREA/a_bound)
            #r_middle = self.R*(self.TANK_AREA/a_middle)
            #r_thermocline = self.THERMOCLINE_MAGIC*a_thermocline 
            
            r_tank = []
            r_thermocline = []
            
            for i in range(self.nodes):
                r_thermocline.append(self.THERMOCLINE_MAGIC*a_thermocline )
                if(i == 0 or i == self.nodes - 1):
                    r_tank.append(self.R*(self.TANK_AREA/a_bound))
                else:
                    r_tank.append(self.R*(self.TANK_AREA/a_middle))
        #print('r_tank:', r_tank)
        #print('r_thermocline:', r_thermocline)
        return r_tank, r_thermocline
    # -------------------- Newton Raphson calculations for Horz ------------------------
    # See Philip's thesis "Rethinking Electric water Heaters"
    
    def getNodeAngle(self, iter_volume):
        stop_threshold = math.pi/100000;
        
        # -------- Raphson -----------------
        theta = math.pi #Initial guess
        while(abs(self.nrf(theta, iter_volume)) > stop_threshold):
            theta = theta - (self.nrf(theta, iter_volume)/self.nrf_prime(theta))
        # ----------------------------------
        return theta
    
    
    def nrf(self, theta, iter_volume):
        volume_m3 = iter_volume*0.001
        return ((2*volume_m3)/(math.pow(self.TANK_RADIUS, 2)*self.TANK_LENGTH)) - theta + math.sin(theta)
    
    def nrf_prime(self, theta):
        return -1 + math.cos(theta)
    # --------------------------------------------------------------------------
    
